#include "timer.h"
#include <sys/time.h>
#include <queue>

#ifdef __MINGW32__
#include <sys/timeb.h>
#endif

#ifdef __GNUC__
#define TRUE 1
#define FALSE 0
#endif

//------------------------ Timer * comparison --------------------

struct CompareTimer : public std::binary_function<Timer*, Timer*, bool> {
  bool operator()(const Timer* lhs, const Timer* rhs) const {
    return *lhs < *rhs;
  }
};

//------------------------ TimerThread ---------------------------

class TimerThread : public Thread {
    private:
        static TimerThread* singleton;
        static Synchronize* p_sync;
        priority_queue<Timer*, vector<Timer*>, CompareTimer> queue;
        Monitor monitor;
        TimerThread() : Thread(NULL) {};
        TimerThread(const TimerThread&){};
        void operator=(const TimerThread&){};
        ~TimerThread() {};
    public:
        static TimerThread *getInstance();
        void run (void *args=NULL);
        void enque(Timer *task);
        void erase(Timer *task);
};

TimerThread* TimerThread::singleton = NULL;
Synchronize* TimerThread::p_sync = new Synchronize();

TimerThread *TimerThread::getInstance() {
    if (singleton == NULL){
        singleton = new TimerThread();
        singleton->start();
    }
    return singleton;
}

void TimerThread::run(void *args) {
    for (;;) {
        if (queue.empty()) monitor.wait();
        else {
            Timer *timer = queue.top();
            if (monitor.wait(timer->getTiming()) == NULL) {
                p_sync->lock();
                queue.pop();
                p_sync->unlock();
                if (timer->isActive()) {
//                  cout << "timer " << timer.tid() << " ticked/stopped" << endl;
                    timer->trigger();
                    if (timer->isCyclic()) {
                        //cout << timer->tid() << " second " << timer->getTiming()->tv_sec << " nano " << timer->getTiming()->tv_nsec << endl; 
                        Timer::inc_time(timer->getTiming(),timer->getPeriod());
//                      cout << "timer " << timer.tid() << " started again" << endl;
                        p_sync->lock();
                        queue.push(timer);
                        p_sync->unlock();
                    }
                }
            }
        }
    }
}

void TimerThread::enque(Timer *task) {
    p_sync->lock();
    queue.push(task);
    p_sync->unlock();
    monitor.notify(this);
}

//------------------------- Timer ----------------------------

unsigned int Timer::nid = 0;

Timer::Timer(Proxy& proxy) {
    _nid = ++nid;
    this->proxy = &proxy;
    tim.tv_sec = 0;
    tim.tv_nsec = 0;
    period = 0;
    active = false;
//    cout << "timer " << _nid << " created" << endl;
}

unsigned int Timer::tid() {
    return _nid;
}

struct timespec *Timer::getTiming() {
    return &tim;
}

bool Timer::isCyclic() {
    return (period != 0);
}

bool Timer::isActive() {
    return active;
}

int Timer::getPeriod() {
    return period;
}

void Timer::start(int ms_offset, int ms_period, int round_ms) {
    get_time(&tim);
    //cout << "Tu som \n" << endl;
    if (round_ms > 0) {
        int round_ns = (round_ms % 1000) * 1000000;
        int round_s = round_ms / 1000;
        if (round_ns > 1) {
                tim.tv_nsec += (round_ns - 1);
                tim.tv_nsec -= (tim.tv_nsec % round_ns);
        }
        if (round_s >= 1) {
                tim.tv_nsec = 0;
                tim.tv_sec += (round_s - 1);
                tim.tv_sec -= (tim.tv_sec % round_s);
        }
    }
    inc_time(&tim,ms_offset);
    period = ms_period;
    active = true;
    //cout << "timer " << _nid << " started" << endl;
    TimerThread::getInstance()->enque(this);
}

void Timer::start(struct timespec *abstime, int ms_period) {
    tim.tv_sec = abstime->tv_sec;
    tim.tv_nsec = abstime->tv_nsec;
    period = ms_period;
    active = true;
    cout << "timer " << _nid << " started" << endl;
    TimerThread::getInstance()->enque(this);
}

void Timer::stop() {
//    tim.tv_sec = 0;
//    tim.tv_nsec = 0;
//    period = 0;
    active = false;
}

void Timer::trigger() {
    proxy->trigger();
}

void Timer::get_time (struct timespec *now) {
#ifdef _TIMEB_H_
	struct timeb tim;
	ftime(&tim);
	now->tv_sec = tim.time;
	now->tv_nsec = tim.millitm*1000000;
#else
	clock_gettime(CLOCK_REALTIME,now);
#endif
}

void Timer::inc_time (struct timespec *now, int ms) {
    int pom = (ms % 1000) * 1000000 + now->tv_nsec;
    if (pom >= 1000000000) {
        now->tv_sec = now->tv_sec + (ms / 1000) + 1;
        now->tv_nsec = pom - 1000000000;
    }
    else {
        now->tv_sec = now->tv_sec + (ms / 1000);
        now->tv_nsec = pom;
    }
}

bool Timer::cmp_time (struct timespec *tim, struct timespec *limit) {
    if (tim->tv_sec < limit->tv_sec) return TRUE;
    else if (tim->tv_sec == limit->tv_sec && tim->tv_nsec <= limit->tv_nsec) return TRUE;
    else return NULL;
}

bool Timer::operator<(const Timer& rhs) const {
    return (
        this->tim.tv_sec > rhs.tim.tv_sec ||
        (this->tim.tv_sec == rhs.tim.tv_sec &&
         (this->tim.tv_nsec > rhs.tim.tv_nsec ||
          (this->tim.tv_nsec == rhs.tim.tv_nsec &&
           this->nid < rhs.nid)))
    );
}

bool Timer::operator==(const Timer& rhs) const {
    return (
        this->tim.tv_sec == rhs.tim.tv_sec &&
        this->tim.tv_nsec == rhs.tim.tv_nsec &&
        this->nid == rhs.nid
    );
}

ostream& operator<<(ostream &os,Timer &timer) {
    os << "Timer " << timer.tid() << " " << timer.getTiming()->tv_sec
       << " : " << timer.getTiming()->tv_nsec;
    return os;
}
