#include "agentspace/agentspace.h"

#include <QtGui/QApplication>
#include "headcontrolgui.h"
#include "armcontrolgui.h"
#include "worldgui.h"

#include <yarp/os/all.h>
#include <yarp/sig/all.h>
#include <yarp/dev/all.h>

#include <string>
#include <iostream>
#include <sstream>

/**
 * \brief Implemetation of agents for actuators and sensors of iCub simulator
 *
 * This classes provides the means of building iCub control system in terms of Agent-Space architecture
 * \addtogroup A iCub_Agent-Space
 *  @{
 */

/******************** Config Block ********************/
/**
 * \addtogroup B Configuration Blocks
 *
 * Structures of configuration block of actuator and sensor agents
 *  @{
 */

/**
 * \typedef ConfBlock
 * The structure of configuration block of MotorAgent
 */
typedef struct {
    string agentName;               /**< the name of Agent */
    string jointBlock;              /**< the name of block for joints angle */
    string speedBlock;              /**< the name of block for joints speed */
    string accelerationBlock;       /**< the name of block for joints acceleration */
    string yarpDevice;              /**< the name of YARP device port*/
    string yarpLocal;               /**< the name of YARP local port */
    string yarpRemote;              /**< the name of YARP remote port */

} ConfBlock;

/**
 * \typedef VelocityBlock
 * The structure of configuration block of VelocityMotorAgent
 */
typedef struct {
    string agentName;               /**< the name of Agent */
    string velocityBlock;           /**< the name of block for joints velocity */
    string encodersBlock;           /**< the name of block for joints encoder */
    string yarpDevice;              /**< the name of YARP device port*/
    string yarpLocal;               /**< the name of YARP local port*/
    string yarpRemote;              /**< the name of YARP remote port*/

} VelocityBlock;

/**
 * \typedef PositionBlock
 * The structure of configuration block of PositionAgent
 */
typedef struct {
    string agentName;               /**< the name of Agent */
    string encodersBlock;           /**< the name of block for joints position */
    string yarpDevice;              /**< the name of YARP device port*/
    string yarpLocal;               /**< the name of YARP local port*/
    string yarpRemote;              /**< the name of YARP remote port*/
    int startTime;                  /**< the start time in ms, when the timer should start */
    int period;                     /**< the period of timer in ms  */

} PositionBlock;

/**
 * \typedef SensorBlock
 * The structure of configuration block of Sensor Agent
 */
typedef struct {
    string agentName;               /**< the name of Agent */
    string portName;                /**< the name of YARP local port */
    string deviceName;              /**< the name of YARP sensor port */
    string outBlock;                /**< the name of block for sensor data */
    string message;                 /**< the message for passing to sensor if needed */
    int startTime;                  /**< the start time in ms, when the timer should start */
    int period;                     /**< the period of timer in ms  */

} SensorBlock;

/**
 * \typedef AgentBlock
 * The basic structure of configuration block of Agent
 */
typedef struct {
    string agentName;               /**< the name of Agent */
    string inBlock;                 /**< the name of block for data reading  */
    string outBlock;                /**< the name of block for data writing  */
    string posBlock;                 /**< the name of block for joints position */
    int startTime;                  /**< the start time in ms, when the timer should start */
    int period;                     /**< the period of timer in ms  */
} AgentBlock;
/** @}*/

/**
 * \addtogroup C iCub Actuators Agents
 *
 * Set of agents for representig actuators of iCub simulator in terms of Agent-Space archtecture
 *  @{
 */

/******************** Motor Agent ********************/

/**
 * \class MotorAgent
 * This class represent agent that controls the actuator subsystem of iCub motor interface with position control
 */
class MotorAgent : public Agent {
    private:
        yarp::dev::PolyDriver driver;
        yarp::dev::IPositionControl *pos;
        yarp::dev::IVelocityControl *vel;
        yarp::dev::IEncoders *encs;
        yarp::dev::IControlLimits *iLim;
        yarp::sig::Vector encoders;
        yarp::sig::Vector commands;
        yarp::sig::Vector speeds;
        yarp::sig::Vector accelerations;
        yarp::sig::Vector def_commands;
        yarp::sig::Vector def_speeds;
        ConfBlock conf;
        double min;
        double max;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from block and sends it to actuators of iCub
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param confBlock name of configuration block
         */
        MotorAgent(string confBlock);
        ~MotorAgent();
};

/******************** Velovity Motor Agent ********************/
/**
 * \class VelocityMotorAgent
 * This class represent agent that controls the actuator subsystem of iCub motor interface with velocity control
 */

class VelocityMotorAgent : public Agent {
    private:
        yarp::dev::PolyDriver driver;
        yarp::dev::IVelocityControl *vel;
        yarp::dev::IEncoders *encs;
        yarp::sig::Vector encoders;
        yarp::sig::Vector commands;
        yarp::sig::Vector def_commands;
        VelocityBlock conf;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from block and sends it to actuators of iCub
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param velocityBlock name of configuration block
         */
        VelocityMotorAgent(string velocityBlock);
        ~VelocityMotorAgent();
};

/******************** Position Control Agent ********************/
/**
 * \class PositionAgent
 * This class represent agent that gets the actual angle of joints
 */
class PositionAgent : public Agent {
    private:
        yarp::dev::PolyDriver driver;
        yarp::dev::IEncoders *encs;
        yarp::sig::Vector encoders;
        PositionBlock conf;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from iCub and writes them to block in Space
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param posBlock name of configuration block
         */
        PositionAgent(string posBlock);
        ~PositionAgent();
};
/** @}*/

/**
 * \addtogroup D iCub Sensor Agents
 *
 * Set of agents for representig sensors of iCub simulator in terms of Agent-Space archtecture
 *  @{
 */

/******************** Camera Senzor Agent ********************/
/**
 * \class CamSensorAgent
 * This class represent agent that gets images from iCub camera
 */
class CamSensorAgent : public Agent {
    private:
        yarp::os::BufferedPort<yarp::sig::ImageOf<yarp::sig::PixelRgb> > *imagePort;
        SensorBlock conf;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from iCub and writes them to block in Space
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param sensorBlock name of configuration block
         */
        CamSensorAgent(string sensorBlock);
        ~CamSensorAgent();
};

/******************** Hand Position Agent ********************/
/**
 * \class HandPosSensorAgent
 * This class represent agent that gets position of iCub hand in world
 */
class HandPosSensorAgent : public Agent {
    private:
        yarp::os::Bottle reply;
        yarp::os::Port *port;
        yarp::sig::Vector position;
        SensorBlock conf;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from iCub and writes them to block in Space
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param sensorBlock name of configuration block
         */
        HandPosSensorAgent(string sensorBlock);
        ~HandPosSensorAgent();
};

/******************** Intertial Senzor Agent ********************/
/**
 * \class InertialSensorAgent
 * This class represent agent that gets data from inertial sensors of iCub
 */
class InertialSensorAgent : public Agent {
    private:
        yarp::os::BufferedPort<yarp::sig::Vector> *port;
        yarp::sig::Vector respond;
        SensorBlock conf;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from iCub and writes them to block in Space
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param sensorBlock name of configuration block
         */
        InertialSensorAgent(string sensorBlock);
        ~InertialSensorAgent();
};

/******************** Touch Senzor Agent ********************/
/**
 * \class TouchSensorAgent
 * This class represent agent that gets data touch/pressure sensors of iCub
 */
class TouchSensorAgent : public Agent {
    private:
        yarp::os::BufferedPort<yarp::sig::Vector> *port;
        yarp::sig::Vector respond;
        SensorBlock conf;

    protected:
        /// Called by constructor of Agent class. Reads configuration block and sets variables
        void init (string args);
        /// Reads the data from iCub and writes them to block in Space
        void sense_select_act(int pid);
     public:
        /**
         * Constructor of Agent
         * \param sensorBlock name of configuration block
         */
        TouchSensorAgent(string sensorBlock);
        ~TouchSensorAgent();
};
/** @}*/

/** @}*/
